from functools import wraps


def filter(filter_function):
    def filter_section_bears(bears, args):
        return {section:
                tuple(bear for bear in bears[section]
                      if filter_function(bear, args))
                for section in bears}

    @wraps(filter_function)
    def filter_wrapper(section_to_bears_dict, args):
        args = {arg.lower() for arg in args}
        local_bears, global_bears = section_to_bears_dict
        local_bears = filter_section_bears(local_bears, args)
        global_bears = filter_section_bears(global_bears, args)
        return local_bears, global_bears

    return filter_wrapper



from coalib.bearlib.languages.documentation.DocstyleDefinition import (
    DocstyleDefinition)
from coalib.results.Diff import Diff
from coalib.results.TextRange import TextRange
from coalib.bearlib.languages.documentation.DocumentationExtraction import (
    extract_documentation_with_markers)


class DocBaseClass:
    """
    DocBaseClass holds important functions which will extract, parse
    and generates diffs for documentation. All bears that processes
    documentation should inherit from this.
    """

    @staticmethod
    def extract(content, language, docstyle):
        """
        Extracts all documentation texts inside the given source-code-string
        using the coala docstyle definition files.

        The documentation texts are sorted by their order appearing in
        ``content``.

        For more information about how documentation comments are
        identified and extracted, see DocstyleDefinition.doctypes enumeration.

        :param content:            The source-code-string where to extract
                                   documentation from. Needs to be a list
                                   or tuple where each string item is a
                                   single line(including ending whitespaces
                                   like ``\\n``).
        :param language:           The programming language used.
        :param docstyle:           The documentation style/tool used
                                   (e.g. doxygen).
        :raises FileNotFoundError: Raised when the docstyle definition file
                                   was not found.
        :raises KeyError:          Raised when the given language is not
                                   defined in given docstyle.
        :raises ValueError:        Raised when a docstyle definition setting
                                   has an invalid format.
        :return:                   An iterator returning instances of
                                   DocumentationComment or MalformedComment
                                   found in the content.
        """
        docstyle_definition = DocstyleDefinition.load(language, docstyle)
        return extract_documentation_with_markers(
                        content, docstyle_definition)

    @staticmethod
    def generate_diff(file, doc_comment, new_comment):
        """
        Generates diff between the original doc_comment and its fix
        new_comment which are instances of DocumentationComment.

        :param doc_comment:
            Original instance of DocumentationComment.
        :param new_comment:
            Fixed instance of DocumentationComment.
        :return:
            Diff instance.
        """
        diff = Diff(file)

        # We need to update old comment positions, as `assemble()`
        # prepends indentation for first line.
        old_range = TextRange.from_values(
            doc_comment.range.start.line,
            1,
            doc_comment.range.end.line,
            doc_comment.range.end.column)

        # Clearing cached assemble() so a fresh one is fetched.
        new_comment.assemble.cache_clear()

        diff.replace(old_range, new_comment.assemble())
        return diff

    def process_documentation(self, *args, **kwargs):
        """
        Checks and handles the fixing part of documentation.

        :return:
            A tuple of processed documentation and warning_desc.
        """
        raise NotImplementedError('This function has to be implemented for a '
                                  'documentation bear.')



from itertools import chain
import os
from types import MappingProxyType

from pyprint.ClosableObject import ClosableObject

from coala_utils.string_processing import escape
from coalib.settings.Section import Section


class ConfWriter(ClosableObject):

    def __init__(self,
                 file_name,
                 key_value_delimiters=('=',),
                 comment_separators=('#',),
                 key_delimiters=(',', ' '),
                 section_name_surroundings=MappingProxyType({'[': ']'}),
                 section_override_delimiters=('.',),
                 unsavable_keys=('save',),
                 key_value_append_delimiters=('+=',)):
        ClosableObject.__init__(self)

        self.__file_name = file_name
        if (os.path.isdir(self.__file_name)):
            self.__file_name = os.path.join(self.__file_name, '.coafile')

        self.__file = open(self.__file_name, 'w')
        self.__key_value_delimiters = key_value_delimiters
        self.__key_value_append_delimiters = key_value_append_delimiters
        self.__comment_separators = comment_separators
        self.__key_delimiters = key_delimiters
        self.__section_name_surroundings = section_name_surroundings
        self.__section_override_delimiters = section_override_delimiters
        self.__unsavable_keys = unsavable_keys
        self.__closed = False

        self.__key_delimiter = self.__key_delimiters[0]
        self.__key_value_delimiter = key_value_delimiters[0]
        self.__key_value_append_delimiter = key_value_append_delimiters[0]
        (self.__section_name_surrounding_beg,
         self.__section_name_surrounding_end) = (
            tuple(self.__section_name_surroundings.items())[0])

    def _close(self):
        self.__file.close()

    def write_sections(self, sections):
        assert not self.__closed

        for section in sections:
            self.write_section(sections[section])

    def write_section(self, section):
        assert not self.__closed

        if not isinstance(section, Section):
            raise TypeError

        self.__write_section_name(section.name)

        keys = []
        val = None
        section_iter = section.__iter__(ignore_defaults=True)
        try:
            while True:
                setting = section[next(section_iter)]
                if (str(setting) == val and
                    not self.is_comment(setting.key) and
                    (
                        (setting.key not in self.__unsavable_keys) or
                        (not setting.from_cli))):
                    keys.append(setting.key)
                elif ((setting.key not in self.__unsavable_keys) or
                      (not setting.from_cli)):
                    self.__write_key_val(keys, val, section.defaults)
                    keys = [setting.key]
                    val = str(setting)
        except StopIteration:
            self.__write_key_val(keys, val, section.defaults)

    def __write_section_name(self, name):
        assert not self.__closed

        self.__file.write(self.__section_name_surrounding_beg + name +
                          self.__section_name_surrounding_end + '\n')

    def __write_key_val(self, keys, val, defaults):
        assert not self.__closed

        if keys == []:
            return

        if all(self.is_comment(key) for key in keys):
            self.__file.write(val + '\n')
            return

        # Add escape characters as appropriate
        keys = [escape(key, chain(['\\'],
                                  self.__key_value_delimiters,
                                  self.__comment_separators,
                                  self.__key_delimiters,
                                  self.__section_override_delimiters))
                for key in keys]
        val = escape(val, chain(['\\'], self.__comment_separators))

        append_keys = []
        other_keys = []
        for key in keys:
            if (defaults and
                    (key in defaults and
                     val.startswith(str(defaults[key])+','))):
                append_keys.append(key)
            else:
                other_keys.append(key)
        self.__write_keys_val_to_file(append_keys, other_keys, val, defaults)

    def __write_keys_val_to_file(self, append_keys, other_keys, val, defaults):
        """
        This method helps in grouping keys with common appendable values and
        writing all keys to the file.

        :param append_keys: The keys that have to be written with an append
                            delimiter.
        :param other_keys:  The keys that have to be written with a normal
                            delimiter.
        :param val:         The value to be written for the keys.
        :param defaults:    The defaults of the section the keys and value are
                            to be written to.
        """
        if append_keys:
            append_keys = sorted(append_keys,
                                 key=lambda key: len(defaults[str(key)]),
                                 reverse=True)
            write_keys = []
            def_val = None
            for key in append_keys:
                if str(defaults[key]) == def_val:
                    write_keys.append(key)
                    continue
                else:
                    if write_keys:
                        append_val = self.__get_append_val(val, def_val)
                        self.__write_value(
                            write_keys,
                            append_val,
                            self.__key_value_append_delimiter)
                    write_keys = [key]
                    def_val = str(defaults[key])
            append_val = self.__get_append_val(val, def_val)
            self.__write_value(write_keys,
                               append_val,
                               self.__key_value_append_delimiter)

        if other_keys:
            self.__write_value(other_keys,
                               val,
                               self.__key_value_delimiter)

    def __get_append_val(self, val, def_val):
        def_val_list = def_val.split(self.__key_delimiter)
        append_val = (self.__key_delimiter + ' ').join(
            [v.strip() for v in val.split(
                self.__key_delimiter)
             if v not in def_val_list])
        return append_val

    def __write_value(self, keys, val, delimiter):
        self.__file.write((self.__key_delimiter + ' ').join(keys) +
                          ' ' + delimiter + ' ' +
                          val + '\n')

    @staticmethod
    def is_comment(key):
        return key.lower().startswith('comment')



from coalib.bearlib.languages.Language import Language


@Language
class HTML:
    __qualname__ = 'Hypertext Markup Language'
    versions = 2.0, 3.2, 4.0, 4.01, 5, 5.1

    extensions = '.html', '.htm'
    multiline_comment_delimiters = {'<!--': '-->'}
    string_delimiters = {'"': '"', "'": "'"}
    multiline_string_delimiters = string_delimiters
    encapsulators = {'<': '>'}
    encapsulators.update(string_delimiters)



def mode_normal(console_printer, log_printer, args, debug=False):
    import functools

    from coalib.coala_main import run_coala
    from coalib.output.ConsoleInteraction import (
        acquire_settings, nothing_done,
        print_results, print_section_beginning)

    partial_print_sec_beg = functools.partial(
        print_section_beginning,
        console_printer)
    results, exitcode, _ = run_coala(
        print_results=print_results,
        acquire_settings=acquire_settings,
        print_section_beginning=partial_print_sec_beg,
        nothing_done=nothing_done,
        console_printer=console_printer,
        args=args,
        debug=debug)

    return exitcode


def mode_non_interactive(console_printer, args, debug=False):
    import functools

    from coalib.coala_main import run_coala
    from coalib.output.ConsoleInteraction import (
        print_results_no_input, print_section_beginning)

    partial_print_sec_beg = functools.partial(
        print_section_beginning,
        console_printer)
    results, exitcode, _ = run_coala(
        print_results=print_results_no_input,
        print_section_beginning=partial_print_sec_beg,
        force_show_patch=True,
        console_printer=console_printer,
        args=args,
        debug=debug)

    return exitcode


def mode_json(args, debug=False):
    import json

    from coalib.coala_main import run_coala
    from coalib.misc.DictUtilities import inverse_dicts
    from coalib.misc.Exceptions import get_exitcode
    from coalib.output.Logging import configure_json_logging
    from coalib.output.JSONEncoder import create_json_encoder

    if args.log_json:
        log_stream = configure_json_logging()

    JSONEncoder = create_json_encoder(use_relpath=args.relpath)
    results = []

    if args.show_bears:
        try:
            from coalib.parsing.FilterHelper import FilterHelper

            local_bears, global_bears = FilterHelper.apply_filter(
                'language', args.filter_by_language)
            bears = inverse_dicts(local_bears, global_bears)
            for bear, _ in sorted(bears.items(),
                                  key=lambda bear_tuple:
                                  bear_tuple[0].name):
                results.append(bear)
        except BaseException as exception:  # pylint: disable=broad-except
            return get_exitcode(exception)
    else:
        results, exitcode, _ = run_coala(args=args, debug=debug)

    retval = {'bears': results} if args.show_bears else {'results': results}

    if args.log_json:
        retval['logs'] = [json.loads(line) for line in
                          log_stream.getvalue().splitlines()]

    if args.output:
        filename = str(args.output[0])
        with open(filename, 'w+') as fp:
            json.dump(retval, fp,
                      cls=JSONEncoder,
                      sort_keys=True,
                      indent=2,
                      separators=(',', ': '))
    else:
        print(json.dumps(retval,
                         cls=JSONEncoder,
                         sort_keys=True,
                         indent=2,
                         separators=(',', ': ')))

    return 0 if args.show_bears else exitcode


def mode_format(args, debug=False):
    from coalib.coala_main import run_coala
    from coalib.output.ConsoleInteraction import print_results_formatted

    _, exitcode, _ = run_coala(
            print_results=print_results_formatted, args=args, debug=debug)
    return exitcode



from coalib.output.printers.LogPrinter import LogPrinter

from coalib.parsing.InvalidFilterException import InvalidFilterException

from coalib.parsing.filters.LanguageFilter import language_filter
from coalib.parsing.filters.CanDetectFilter import can_detect_filter
from coalib.parsing.filters.CanFixFilter import can_fix_filter


class FilterHelper:
    available_filters = {'language': language_filter,
                         'can_detect': can_detect_filter,
                         'can_fix': can_fix_filter}

    @classmethod
    def get_all_filters_str(cls, sep=', '):
        return sep.join(sorted(cls.available_filters))

    @classmethod
    def is_valid_filter(cls, filter):
        return filter in cls.available_filters

    @classmethod
    def apply_filter(cls, filter_name, filter_args, all_bears=None):
        if all_bears is None:
            from coalib.settings.ConfigurationGathering import (
                get_all_bears)
            all_bears = get_all_bears(LogPrinter())
        if not cls.is_valid_filter(filter_name):
            raise InvalidFilterException('{!r} is an invalid filter. '
                                         'Available filters: {}'.format(
                                             filter_name,
                                             cls.get_all_filters_str()))
        if not filter_args or len(filter_args) == 0:
            return all_bears
        return cls.available_filters[filter_name](all_bears, filter_args)

    @classmethod
    def apply_filters(cls, filters, bears=None):
        """
        Returns bears after filtering based on ``args``. It returns
        intersection of bears if more than one element is present in ``args``
        list.

        :param filters:
            List of args based on ``bears`` has to be filtered.
            For example:
            ``[['language', 'c', 'java'], ['can_fix', 'syntax']]``
        :param bears:
            The bears to filter.
        :return:
            Filtered bears.
        """
        for filter in filters:
            filter_name, *filter_args = filter
            bears = cls.apply_filter(
                filter_name, filter_args, bears)
        return bears



import uuid
from os.path import relpath

from coala_utils.decorators import (
    enforce_signature, generate_ordering, generate_repr, get_public_members)
from coalib.bearlib.aspects import aspectbase
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.SourceRange import SourceRange


# Omit additional info, debug message and diffs for brevity
@generate_repr(('id', hex),
               'origin',
               'affected_code',
               ('severity', RESULT_SEVERITY.reverse.get),
               'confidence',
               'message',
               ('aspect', lambda aspect: type(aspect).__qualname__),
               'applied_actions')
@generate_ordering('affected_code',
                   'severity',
                   'confidence',
                   'origin',
                   'message_base',
                   'message_arguments',
                   'aspect',
                   'additional_info',
                   'debug_msg',
                   'applied_actions')
class Result:
    """
    A result is anything that has an origin and a message.

    Optionally it might affect a file.

    Result messages can also have arguments. The message is python
    style formatted with these arguments.

    >>> r = Result('origin','{arg1} and {arg2}', \
           message_arguments={'arg1': 'foo', 'arg2': 'bar'})
    >>> r.message
    'foo and bar'

    Message arguments may be changed later. The result message
    will also reflect these changes.

    >>> r.message_arguments = {'arg1': 'spam', 'arg2': 'eggs'}
    >>> r.message
    'spam and eggs'

    """

    @enforce_signature
    def __init__(self,
                 origin,
                 message: str,
                 affected_code: (tuple, list)=(),
                 severity: int=RESULT_SEVERITY.NORMAL,
                 additional_info: str='',
                 debug_msg='',
                 diffs: (dict, None)=None,
                 confidence: int=100,
                 aspect: (aspectbase, None)=None,
                 message_arguments: dict={},
                 applied_actions: dict={}):
        """
        :param origin:
            Class name or creator object of this object.
        :param message:
            Base message to show with this result.
        :param affected_code:
            A tuple of SourceRange objects pointing to related positions in the
            source code.
        :param severity:
            Severity of this result.
        :param additional_info:
            A long description holding additional information about the issue
            and/or how to fix it. You can use this like a manual entry for a
            category of issues.
        :param debug_msg:
            A message which may help the user find out why this result was
            yielded.
        :param diffs:
            A dictionary with filename as key and ``Diff`` object
            associated with it as value.
        :param confidence:
            A number between 0 and 100 describing the likelihood of this result
            being a real issue.
        :param aspect:
            An aspectclass instance which this result is associated to.
            Note that this should be a leaf of the aspect tree!
            (If you have a node, spend some time figuring out which of
            the leafs exactly your result belongs to.)
        :param message_arguments:
            Arguments to be provided to the base message.
        :param applied_actions:
            A dictionary that contains the result, file_dict, file_diff_dict and
            the section for an action.
        :raises ValueError:
            Raised when confidence is not between 0 and 100.
        :raises KeyError:
            Raised when message_base can not be formatted with
            message_arguments.
        """
        origin = origin or ''
        if not isinstance(origin, str):
            origin = origin.__class__.__name__
        if severity not in RESULT_SEVERITY.reverse:
            raise ValueError('severity is not a valid RESULT_SEVERITY')

        self.origin = origin
        self.message_base = message
        self.message_arguments = message_arguments
        self.applied_actions = applied_actions
        if message_arguments:
            self.message_base.format(**self.message_arguments)
        self.debug_msg = debug_msg
        self.additional_info = additional_info
        # Sorting is important for tuple comparison
        self.affected_code = tuple(sorted(affected_code))
        self.severity = severity
        if confidence < 0 or confidence > 100:
            raise ValueError('Value of confidence should be between 0 and 100.')
        self.confidence = confidence
        self.diffs = diffs
        self.id = uuid.uuid4().int
        self.aspect = aspect

    @property
    def message(self):
        if not self.message_arguments:
            return self.message_base
        return self.message_base.format(**self.message_arguments)

    @message.setter
    def message(self, value: str):
        self.message_base = value

    def set_applied_actions(self, applied_actions):
        self.applied_actions = applied_actions

    def get_applied_actions(self):
        return self.applied_actions

    @classmethod
    @enforce_signature
    def from_values(cls,
                    origin,
                    message: str,
                    file: str,
                    line: (int, None)=None,
                    column: (int, None)=None,
                    end_line: (int, None)=None,
                    end_column: (int, None)=None,
                    severity: int=RESULT_SEVERITY.NORMAL,
                    additional_info: str='',
                    debug_msg='',
                    diffs: (dict, None)=None,
                    confidence: int=100,
                    aspect: (aspectbase, None)=None,
                    message_arguments: dict={}):
        """
        Creates a result with only one SourceRange with the given start and end
        locations.

        :param origin:
            Class name or creator object of this object.
        :param message:
            Base message to show with this result.
        :param message_arguments:
            Arguments to be provided to the base message
        :param file:
            The related file.
        :param line:
            The first related line in the file. (First line is 1)
        :param column:
            The column indicating the first character. (First character is 1)
        :param end_line:
            The last related line in the file.
        :param end_column:
            The column indicating the last character.
        :param severity:
            Severity of this result.
        :param additional_info:
            A long description holding additional information about the issue
            and/or how to fix it. You can use this like a manual entry for a
            category of issues.
        :param debug_msg:
            A message which may help the user find out why this result was
            yielded.
        :param diffs:
            A dictionary with filename as key and ``Diff`` object
            associated with it as value.
        :param confidence:
            A number between 0 and 100 describing the likelihood of this result
            being a real issue.
        :param aspect:
            An Aspect object which this result is associated to. Note that this
            should be a leaf of the aspect tree! (If you have a node, spend
            some time figuring out which of the leafs exactly your result
            belongs to.)
        """
        range = SourceRange.from_values(file,
                                        line,
                                        column,
                                        end_line,
                                        end_column)

        return cls(origin=origin,
                   message=message,
                   affected_code=(range,),
                   severity=severity,
                   additional_info=additional_info,
                   debug_msg=debug_msg,
                   diffs=diffs,
                   confidence=confidence,
                   aspect=aspect,
                   message_arguments=message_arguments)

    def to_string_dict(self):
        """
        Makes a dictionary which has all keys and values as strings and
        contains all the data that the base Result has.

        FIXME: diffs are not serialized ATM.
        FIXME: Only the first SourceRange of affected_code is serialized. If
        there are more, this data is currently missing.

        :return: Dictionary with keys and values as string.
        """
        retval = {}

        members = ['id',
                   'additional_info',
                   'debug_msg',
                   'message',
                   'message_base',
                   'message_arguments',
                   'origin',
                   'confidence']

        for member in members:
            value = getattr(self, member)
            retval[member] = '' if value is None else str(value)

        retval['severity'] = str(RESULT_SEVERITY.reverse.get(
            self.severity, ''))
        if len(self.affected_code) > 0:
            retval['file'] = self.affected_code[0].file
            line = self.affected_code[0].start.line
            retval['line_nr'] = '' if line is None else str(line)
        else:
            retval['file'], retval['line_nr'] = '', ''

        return retval

    @enforce_signature
    def apply(self, file_dict: dict):
        """
        Applies all contained diffs to the given file_dict. This operation will
        be done in-place.

        :param file_dict: A dictionary containing all files with filename as
                          key and all lines a value. Will be modified.
        """
        for filename, diff in self.diffs.items():
            file_dict[filename] = diff.modified

    def __add__(self, other):
        """
        Joins those patches to one patch.

        :param other: The other patch.
        """
        assert isinstance(self.diffs, dict)
        assert isinstance(other.diffs, dict)

        for filename in other.diffs:
            if filename in self.diffs:
                self.diffs[filename] += other.diffs[filename]
            else:
                self.diffs[filename] = other.diffs[filename]

        return self

    def overlaps(self, ranges):
        """
        Determines if the result overlaps with source ranges provided.

        :param ranges: A list SourceRange objects to check for overlap.
        :return:       True if the ranges overlap with the result.
        """
        if isinstance(ranges, SourceRange):
            ranges = [ranges]

        for range in ranges:
            for self_range in self.affected_code:
                if range.overlaps(self_range):
                    return True

        return False

    def location_repr(self):
        """
        Retrieves a string, that briefly represents
        the affected code of the result.

        :return: A string containing all of the affected files
                 separated by a comma.
        """

        if not self.affected_code:
            return 'the whole project'

        # Set important to exclude duplicate file names
        range_paths = set(sourcerange.file
                          for sourcerange in self.affected_code)

        return ', '.join(repr(relpath(range_path))
                         for range_path in sorted(range_paths))

    def __json__(self, use_relpath=False):
        _dict = get_public_members(self)
        if use_relpath and _dict['diffs']:
            _dict['diffs'] = {relpath(file): diff
                              for file, diff in _dict['diffs'].items()}
        _dict['aspect'] = type(self.aspect).__qualname__
        return _dict



import copy
import difflib
import logging

from unidiff import PatchSet

from coalib.results.LineDiff import LineDiff, ConflictError
from coalib.results.SourceRange import SourceRange
from coalib.results.TextRange import TextRange
from coala_utils.decorators import enforce_signature, generate_eq


@generate_eq('_file', 'modified', 'rename', 'delete')
class Diff:
    """
    A Diff result represents a difference for one file.
    """

    def __init__(self, file_list, rename=False, delete=False):
        """
        Creates an empty diff for the given file.

        :param file_list: The original (unmodified) file as a list of its
                          lines.
        :param rename:    False or str containing new name of file.
        :param delete:    True if file is set to be deleted.
        """
        self._changes = {}
        self._file = list(file_list)
        self._original = self._generate_linebreaks(self._file)
        self.rename = rename
        self.delete = delete

    @classmethod
    def from_string_arrays(cls, file_array_1, file_array_2, rename=False):
        """
        Creates a Diff object from two arrays containing strings.

        If this Diff is applied to the original array, the second array will be
        created.

        :param file_array_1: Original array
        :param file_array_2: Array to compare
        :param rename:       False or str containing new name of file.
        """
        result = cls(file_array_1, rename=rename)

        matcher = difflib.SequenceMatcher(None, file_array_1, file_array_2)
        # We use this because its faster (generator) and doesn't yield as much
        # useless information as get_opcodes.
        for change_group in matcher.get_grouped_opcodes(1):
            for (tag,
                 a_index_1,
                 a_index_2,
                 b_index_1,
                 b_index_2) in change_group:
                if tag == 'delete':
                    for index in range(a_index_1+1, a_index_2+1):
                        result.delete_line(index)
                elif tag == 'insert':
                    # We add after line, they add before, so dont add 1 here
                    result.add_lines(a_index_1,
                                     file_array_2[b_index_1:b_index_2])
                elif tag == 'replace':
                    result.modify_line(a_index_1+1,
                                       file_array_2[b_index_1])
                    result.add_lines(a_index_1+1,
                                     file_array_2[b_index_1+1:b_index_2])
                    for index in range(a_index_1+2, a_index_2+1):
                        result.delete_line(index)

        return result

    @classmethod
    def from_unified_diff(cls, unified_diff, original_file):
        """
        Creates a ``Diff`` object from given unified diff.

        If the provided unified diff does not contain any patch,
        the ``Diff`` object initialized from the original file is
        returned.

        :param unified_diff:  Unified diff string.
        :param original_file: The contents of the original file
                              (line-splitted).
        :raises RuntimeError: Raised when the context lines or the
                              lines to be removed do not match in
                              the original file and the unified diff.
        """
        patch_set = PatchSet(unified_diff.splitlines())

        diff = Diff(original_file)

        if not patch_set:
            return diff

        # FIXME Handle patches consisting of changes in more than one file
        file = patch_set[0]

        for hunk in file:
            file_line = hunk.source_start
            hunk_iterator = iter(hunk)

            try:
                while True:
                    line = next(hunk_iterator)
                    source_code = str(line)[1:]
                    if line.is_added:
                        add_set = []
                        # As ``Diff`` does not allow line additions to a
                        # position more than one time, add all the
                        # consecutive '+' lines at once.
                        try:
                            while line.is_added:
                                add_set.append(source_code)
                                line = next(hunk_iterator)
                                source_code = str(line)[1:]

                            diff.add_lines(file_line-1, add_set)

                        except StopIteration:
                            diff.add_lines(file_line-1, add_set)
                            break

                    original_line = original_file[file_line-1].rstrip('\n')

                    if line.is_removed:
                        if source_code != original_line:
                            raise RuntimeError(
                                'The line to delete does not match with '
                                'the line in the original file. '
                                'Line to delete: {!r}, '
                                'Original line #{!r}: {!r}'.format(
                                    source_code,
                                    file_line,
                                    original_line)
                                )
                        diff.delete_line(file_line)

                    else:
                        if source_code != original_line:
                            raise RuntimeError(
                                'Context lines do not match. '
                                'Line from unified diff: {!r}, '
                                'Original line #{!r}: {!r}'.format(
                                    source_code,
                                    file_line,
                                    original_line)
                                )

                    file_line += 1

            except StopIteration:
                pass

        return diff

    @classmethod
    def from_clang_fixit(cls, fixit, file):
        """
        Creates a Diff object from a given clang fixit and the file contents.

        :param fixit: A cindex.Fixit object.
        :param file:  A list of lines in the file to apply the fixit to.
        :return:      The corresponding Diff object.
        """
        assert isinstance(file, (list, tuple))

        oldvalue = '\n'.join(file[fixit.range.start.line-1:
                                  fixit.range.end.line])
        endindex = fixit.range.end.column - len(file[fixit.range.end.line-1])-1

        newvalue = (oldvalue[:fixit.range.start.column-1] +
                    fixit.value +
                    oldvalue[endindex:])
        new_file = (file[:fixit.range.start.line-1] +
                    type(file)(newvalue.splitlines(True)) +
                    file[fixit.range.end.line:])

        return cls.from_string_arrays(file, new_file)

    def _get_change(self, line_nr, min_line=1):
        if not isinstance(line_nr, int):
            raise TypeError('line_nr needs to be an integer.')
        if line_nr < min_line:
            raise IndexError('The given line number is not allowed.')

        return self._changes.get(line_nr, LineDiff())

    def stats(self):
        """
        Returns tuple containing number of additions and deletions in the diff.
        """
        additions = 0
        deletions = 0
        for line_diff in self._changes.values():
            if line_diff.change:
                additions += 1
                deletions += 1
            elif line_diff.delete:
                deletions += 1
            if line_diff.add_after:
                additions += len(line_diff.add_after)
        return additions, deletions

    def __len__(self):
        """
        Returns total number of additions and deletions in diff.
        """
        return sum(self.stats())

    @property
    def rename(self):
        """
        :return: string containing new name of the file.
        """
        return self._rename

    @rename.setter
    @enforce_signature
    def rename(self, rename: (str, False)):
        """
        :param rename: False or string containing new name of file.
        """
        self._rename = rename

    @property
    def delete(self):
        """
        :return: True if file is set to be deleted.
        """
        return self._delete

    @delete.setter
    @enforce_signature
    def delete(self, delete: bool):
        """
        :param delete: True if file is set to be deleted, False otherwise.
        """
        self._delete = delete

    @property
    def original(self):
        """
        Retrieves the original file.
        """
        return self._original

    def _raw_modified(self):
        """
        Calculates the modified file, after applying the Diff to the original.
        """
        result = []

        if self.delete:
            return result

        current_line = 0

        # Note that line_nr counts from _1_ although 0 is possible when
        # inserting lines before everything
        for line_nr in sorted(self._changes):
            result.extend(self._file[current_line:max(line_nr-1, 0)])
            linediff = self._changes[line_nr]
            if not linediff.delete and not linediff.change and line_nr > 0:
                result.append(self._file[line_nr-1])
            elif linediff.change:
                result.append(linediff.change[1])

            if linediff.add_after:
                result.extend(linediff.add_after)

            current_line = line_nr

        result.extend(self._file[current_line:])

        return result

    @property
    def modified(self):
        """
        Calculates the modified file, after applying the Diff to the original.

        This property also adds linebreaks at the end of each line.
        If no newline was present at the end of file before, this state will
        be preserved, except if the last line is deleted.
        """
        return self._generate_linebreaks(self._raw_modified())

    @property
    def unified_diff(self):
        """
        Generates a unified diff corresponding to this patch.

        Each change will be displayed on its own line. Additionally, the
        unified diff preserves the EOF-state of the original file. This
        means that the ``Diff`` will only have a linebreak on the last line,
        if that was also present in the original file.

        Note that the unified diff is not deterministic and thus not suitable
        for equality comparison.
        """

        list_unified_diff = list(difflib.unified_diff(
            self._file,
            self._raw_modified(),
            tofile=self.rename if isinstance(self.rename, str) else ''))

        return ''.join(self._generate_linebreaks(list_unified_diff))

    def __json__(self):
        """
        Override JSON export, using the unified diff is the easiest thing for
        the users.
        """
        return self.unified_diff

    def affected_code(self, filename):
        """
        Creates a list of SourceRange objects which point to the related code.
        Changes on continuous lines will be put into one SourceRange.

        :param filename: The filename to associate the SourceRange's to.
        :return:         A list of all related SourceRange objects.
        """
        return list(diff.range(filename)
                    for diff in self.split_diff(distance=0))

    def split_diff(self, distance=1):
        """
        Splits this diff into small pieces, such that several continuously
        altered lines are still together in one diff. All subdiffs will be
        yielded.

        A diff like this with changes being together closely won't be splitted:

        >>> diff = Diff.from_string_arrays([     'b', 'c', 'e'],
        ...                                ['a', 'b', 'd', 'f'])
        >>> len(list(diff.split_diff()))
        1

        If we set the distance to 0, it will be splitted:

        >>> len(list(diff.split_diff(distance=0)))
        2

        If a negative distance is given, every change will be yielded as an own
        diff, even if they are right beneath each other:

        >>> len(list(diff.split_diff(distance=-1)))
        3

        If a file gets renamed or deleted only, it will be yielded as is:

        >>> len(list(Diff([], rename='test').split_diff()))
        1

        An empty diff will not yield any diffs:

        >>> len(list(Diff([]).split_diff()))
        0

        :param distance: Number of unchanged lines that are allowed in between
                         two changed lines so they get yielded as one diff.
        """
        if not self:
            return

        last_line = -1
        this_diff = Diff(self._file, rename=self.rename, delete=self.delete)
        for line in sorted(self._changes.keys()):
            if line > last_line + distance + 1 and len(this_diff._changes) > 0:
                yield this_diff
                this_diff = Diff(self._file, rename=self.rename,
                                 delete=self.delete)

            last_line = line
            this_diff._changes[line] = self._changes[line]

        # If the diff contains no line changes, the loop above will not be run
        # else, this_diff will never be empty and thus this has to be yielded
        # always.
        yield this_diff

    def range(self, filename):
        """
        Calculates a SourceRange spanning over the whole Diff. If something is
        added after the 0th line (i.e. before the first line) the first line
        will be included in the SourceRange.

        The range of an empty diff will only affect the filename:

        >>> range = Diff([]).range("file")
        >>> range.file is None
        False
        >>> print(range.start.line)
        None

        :param filename: The filename to associate the SourceRange with.
        :return:         A SourceRange object.
        """
        if len(self._changes) == 0:
            return SourceRange.from_values(filename)

        start = min(self._changes.keys())
        end = max(self._changes.keys())
        return SourceRange.from_values(filename,
                                       start_line=max(1, start),
                                       end_line=max(1, end))

    def __add__(self, other):
        """
        Adds another diff to this one. Will throw an exception if this is not
        possible. (This will *not* be done in place.)
        """
        if not isinstance(other, Diff):
            raise TypeError('Only diffs can be added to a diff.')

        if self.rename != other.rename and False not in (self.rename,
                                                         other.rename):
            raise ConflictError('Diffs contain conflicting renamings.')

        result = copy.deepcopy(self)
        result.rename = self.rename or other.rename
        result.delete = self.delete or other.delete

        for line_nr in other._changes:
            change = other._changes[line_nr]
            if change.delete is True:
                result.delete_line(line_nr)
            if change.add_after is not False:
                result.add_lines(line_nr, change.add_after)
            if change.change is not False:
                result.modify_line(line_nr, change.change[1])

        return result

    def __bool__(self):
        """
        >>> bool(Diff([]))
        False
        >>> bool(Diff([], rename="some"))
        True
        >>> bool(Diff([], delete=True))
        True
        >>> bool(Diff.from_string_arrays(['1'], []))
        True

        :return: False if the patch has no effect at all when applied.
        """
        return (self.rename is not False or
                self.delete is True or
                self.modified != self.original)

    def delete_line(self, line_nr):
        """
        Mark the given line nr as deleted. The first line is line number 1.

        Raises an exception if line number doesn't exist in the diff.
        """
        if line_nr > len(self._file):
            raise IndexError('The given line number is out of bounds.')

        linediff = self._get_change(line_nr)
        linediff.delete = True
        self._changes[line_nr] = linediff

    def delete_lines(self, line_nr_start, line_nr_end):
        """
        Delete lines in a specified range, inclusively.

        The range must be valid, i.e. lines must exist in diff, else an
        exception is raised.
        """
        for line_nr in range(line_nr_start, line_nr_end + 1):
            self.delete_line(line_nr)

    def add_lines(self, line_nr_before, lines):
        """
        Adds lines after the given line number.

        :param line_nr_before: Line number of the line before the additions.
                               Use 0 for insert lines before everything.
        :param lines:          A list of lines to add.
        """
        if lines == []:
            return  # No action

        linediff = self._get_change(line_nr_before, min_line=0)
        if linediff.add_after is not False:
            raise ConflictError('Cannot add lines after the given line since '
                                'there are already lines.')

        linediff.add_after = lines
        self._changes[line_nr_before] = linediff

    def add_line(self, line_nr_before, line):
        """
        Adds line after the given line number.

        :param line_nr_before: Line number of the line before the addition.
                               Use 0 to insert line before everything.
        :param line:           Line to add.
        """
        return self.add_lines(line_nr_before, [line])

    def modify_line(self, line_nr, replacement):
        r"""
        Changes the given line with the given line number. The replacement will
        be there instead.

        Given an empty diff object:

        >>> diff = Diff(['Hey there! Gorgeous.\n',
        ...              "It's nice that we're here.\n"])

        We can change a line easily:

        >>> diff.modify_line(1,
        ...                  'Hey there! This is sad.\n')
        >>> diff.modified
        ['Hey there! This is sad.\n', "It's nice that we're here.\n"]

        We can even merge changes within one line:

        >>> diff.modify_line(1,
        ...                  'Hello. :( Gorgeous.\n')
        >>> diff.modified
        ['Hello. :( This is sad.\n', "It's nice that we're here.\n"]

        However, if we change something that has been changed before, we'll get
        a conflict:

        >>> diff.modify_line(1, 'Hello. This is not ok. Gorgeous.\n')
        Traceback (most recent call last):
         ...
        coalib.results.LineDiff.ConflictError: ...
        """
        linediff = self._get_change(line_nr)
        if linediff.change is not False and linediff.change[1] != replacement:
            if len(replacement) == len(linediff.change[1]) == 1:
                raise ConflictError('Cannot merge the given line changes.')

            # The following diffs are created from strings, instead of lists.
            orig_diff = Diff.from_string_arrays(linediff.change[0],
                                                linediff.change[1])
            new_diff = Diff.from_string_arrays(linediff.change[0],
                                               replacement)
            replacement = ''.join((orig_diff + new_diff)._raw_modified())

        linediff.change = (self._file[line_nr-1], replacement)
        self._changes[line_nr] = linediff

    def change_line(self, line_nr, original_line, replacement):
        logging.debug('Use of change_line method is deprecated. Instead '
                      'use modify_line method, without the original_line '
                      'argument')
        self.modify_line(line_nr, replacement)

    def replace(self, range, replacement):
        r"""
        Replaces a part of text. Allows to span multiple lines.

        This function uses ``add_lines`` and ``delete_lines`` accordingly, so
        calls of those functions on lines given ``range`` affects after usage
        or vice versa lead to ``ConflictError``.

        >>> from coalib.results.TextRange import TextRange
        >>> test_text = ['hello\n', 'world\n', '4lines\n', 'done\n']
        >>> def replace(range, text):
        ...     diff = Diff(test_text)
        ...     diff.replace(range, text)
        ...     return diff.modified
        >>> replace(TextRange.from_values(1, 5, 4, 3), '\nyeah\ncool\nno')
        ['hell\n', 'yeah\n', 'cool\n', 'none\n']
        >>> replace(TextRange.from_values(2, 1, 3, 5), 'b')
        ['hello\n', 'bes\n', 'done\n']
        >>> replace(TextRange.from_values(1, 6, 4, 3), '')
        ['hellone\n']

        :param range:       The ``TextRange`` that gets replaced.
        :param replacement: The replacement string. Can be multiline.
        """
        # Remaining parts of the lines not affected by the replace.
        first_part = (
            self._file[range.start.line - 1][:range.start.column - 1])
        last_part = self._file[range.end.line - 1][range.end.column - 1:]

        self.delete_lines(range.start.line, range.end.line)
        self.add_lines(range.start.line - 1,
                       (first_part + replacement + last_part).splitlines(True))

    def insert(self, position, text):
        r"""
        Inserts (multiline) text at arbitrary position.

        >>> from coalib.results.TextPosition import TextPosition
        >>> test_text = ['123\n', '456\n', '789\n']
        >>> def insert(position, text):
        ...     diff = Diff(test_text)
        ...     diff.insert(position, text)
        ...     return diff.modified
        >>> insert(TextPosition(2, 3), 'woopy doopy')
        ['123\n', '45woopy doopy6\n', '789\n']
        >>> insert(TextPosition(1, 1), 'woopy\ndoopy')
        ['woopy\n', 'doopy123\n', '456\n', '789\n']
        >>> insert(TextPosition(2, 4), '\nwoopy\ndoopy\n')
        ['123\n', '456\n', 'woopy\n', 'doopy\n', '\n', '789\n']

        :param position: The ``TextPosition`` where to insert text.
        :param text:     The text to insert.
        """
        self.replace(TextRange(position, position), text)

    def remove(self, range):
        r"""
        Removes a piece of text in a given range.

        >>> from coalib.results.TextRange import TextRange
        >>> test_text = ['nice\n', 'try\n', 'bro\n']
        >>> def remove(range):
        ...     diff = Diff(test_text)
        ...     diff.remove(range)
        ...     return diff.modified
        >>> remove(TextRange.from_values(1, 1, 1, 4))
        ['e\n', 'try\n', 'bro\n']
        >>> remove(TextRange.from_values(1, 5, 2, 1))
        ['nicetry\n', 'bro\n']
        >>> remove(TextRange.from_values(1, 3, 3, 2))
        ['niro\n']
        >>> remove(TextRange.from_values(2, 1, 2, 1))
        ['nice\n', 'try\n', 'bro\n']

        :param range: The range to delete.
        """
        self.replace(range, '')

    @staticmethod
    def _add_linebreaks(lines):
        """
        Validate that each line in lines ends with a
        newline character and appends one if that is not the case.

        :param lines: A list of strings, representing lines.
        """

        return [line
                if line.endswith('\n')
                else line + '\n'
                for line in lines]

    @staticmethod
    def _generate_linebreaks(lines):
        """
        Validate that each line in lines ends with a
        newline character and appends one if that is not the case.
        Exception is the last line in the list.

        :param lines: A list of strings, representing lines.
        """

        if lines == []:
            return []

        return Diff._add_linebreaks(lines[:-1]) + [lines[-1]]



from coalib.parsing.filters.Filter import filter


@filter
def can_detect_filter(bear, args):
    """
    Filters the bears by ``CAN_DETECT``.

    :param bear: Bear object.
    :param args: Set of detectable issue types on which ``bear`` is to be
                 filtered.
    :return:     ``True`` if this bear matches the criteria inside args,
                 ``False`` otherwise.
    """
    return bool({detect.lower() for detect in bear.CAN_DETECT} & args)



from coalib.bearlib.aspects import Root, Taste


@Root.subaspect
class Spelling:
    """
    How words should be written.
    """
    class docs:
        example = """
        'Tihs si surly som incoreclt speling.
        `Coala` is always written with a lowercase `c`.
        """
        example_language = 'reStructuredText'
        importance_reason = """
        Words should always be written as they are supposed to be;
        standardisation facilitates communication.
        """
        fix_suggestions = """
        Use the correct spelling for the misspelled words.
        """


@Spelling.subaspect
class DictionarySpelling:
    """
    Valid language's words spelling.
    """
    class docs:
        example = """
        This is toatly wonrg.
        """
        example_language = 'reStructuredText'
        importance_reason = """
        Good spelling facilitates communication and avoids confusion. By
        following the same rules for spelling words, we can all understand
        the text we read. Poor spelling distracts the reader and they lose
        focus.
        """
        fix_suggestions = """
        You can use a spell-checker to fix this for you or just ensure
        yourself that things are well written.
        """


@Spelling.subaspect
class OrgSpecificWordSpelling:
    """
    Organisations like coala specified words' spelling.
    """
    class docs:
        example = """
        `Coala` is always written with a lower case c, also at the beginning
        of the sentence.
        """
        example_language = 'reStructuredText'
        importance_reason = """
        There are words you want to be written as you want, like your
        organisation's name.
        """
        fix_suggestions = """
        Simply make sure those words match with what is provided by the
        organisation.
        """
    specific_word = Taste[list](
        'Represents the regex of the specific word to check.',
        (('c[o|O][a|A][l|L][a|A]',), ), default=list())



import copy
import os
import sys
from collections import OrderedDict

from coalib.bearlib.aspects import AspectList
from coalib.collecting.Collectors import collect_registered_bears_dirs
from coala_utils.decorators import enforce_signature, generate_repr
from coalib.misc.DictUtilities import update_ordered_dict_key
from coalib.settings.Setting import Setting, path_list
from coalib.parsing.Globbing import glob_escape


def append_to_sections(sections,
                       key,
                       value,
                       origin,
                       section_name=None,
                       from_cli=False,
                       to_append=False):
    """
    Appends the given data as a Setting to a Section with the given name. If
    the Section does not exist before it will be created empty.

    :param sections:     The sections dictionary to add to.
    :param key:          The key of the setting to add.
    :param value:        The value of the setting to add.
    :param origin:       The origin value of the setting to add.
    :param section_name: The name of the section to add to.
    :param from_cli:     Whether or not this data comes from the CLI.
    :param to_append:    The boolean value if setting value needs to be
                         appended to a setting in the defaults of a section.
    """
    if key == '' or value is None:
        return

    if section_name == '' or section_name is None:
        section_name = 'default'

    if not section_name.lower() in sections:
        sections[section_name.lower()] = Section(section_name)

    sections[section_name.lower()].append(Setting(
        key, str(value), origin, from_cli=from_cli, to_append=to_append))


def extract_aspects_from_section(section):
    """
    Extracts aspects and their related settings from a section and create an
    AspectList from it.

    :param section: Section object.
    :return:        AspectList containing aspectclass instance with
                    user-defined tastes.
    """
    aspects = section.get('aspects')
    language = section.get('language')

    # Skip aspects initialization if not configured in section
    if not len(aspects):
        return None

    if not len(language):
        raise AttributeError('Language was not found in configuration file. '
                             'Usage of aspect-based configuration must '
                             'include language information.')

    aspect_instances = AspectList(exclude=section.get('excludes'))

    for aspect in AspectList(aspects):
        # Search all related tastes in section.
        tastes = {name.split('.')[-1]: value
                  for name, value in section.contents.items()
                  if name.lower().startswith(aspect.__name__.lower())}
        aspect_instances.append(aspect(language, **tastes))

    return aspect_instances


@generate_repr()
class Section:
    """
    This class holds a set of settings.

    To add settings and sections to a dictionary of sections we can use
    ``append_to_sections``:

    >>> sections = {}
    >>> append_to_sections(sections,
    ...                    'test1',
    ...                    'val',
    ...                    'origin',
    ...                    section_name='all')
    >>> 'all' in sections
    True
    >>> len(sections)
    1
    >>> str(sections)
    "{'all': <Section object(contents=OrderedDict([('test1', ..."

    We can also add settings that can be appended to other settings. Basically
    it takes the default value of the setting which resides in the defaults of
    the section and appends the value of the setting in the second and returns
    the value of the setting:

    >>> append_to_sections(sections,
    ...                    'test1',
    ...                    'val2',
    ...                    'origin',
    ...                    section_name='all.python',
    ...                    to_append=True)

    When the section has no defaults:

    >>> str(sections['all.python']['test1'])
    'val2'

    After assigning defaults:

    >>> sections['all.python'].set_default_section(sections)
    >>> str(sections['all.python']['test1'])
    'val, val2'
    """

    @staticmethod
    def __prepare_key(key):
        return str(key).lower().strip()

    def __init__(self,
                 name,
                 defaults=None):
        if defaults is not None and not isinstance(defaults, Section):
            raise TypeError('defaults has to be a Section object or None.')
        if defaults is self:
            raise ValueError('defaults may not be self for non-recursivity.')

        self.name = str(name)
        self.defaults = defaults
        self.contents = OrderedDict()

    def bear_dirs(self):
        bear_dirs = path_list(self.get('bear_dirs', ''))
        for bear_dir in bear_dirs:
            sys.path.append(bear_dir)
        bear_dir_globs = [
            os.path.join(glob_escape(bear_dir), '**')
            for bear_dir in bear_dirs]
        bear_dir_globs += [
            os.path.join(glob_escape(bear_dir), '**')
            for bear_dir in collect_registered_bears_dirs('coalabears')]
        return bear_dir_globs

    def is_enabled(self, targets):
        """
        Checks if this section is enabled or, if targets is not empty, if it is
        included in the targets list.

        :param targets: List of target section names, all lower case.
        :return:        True or False
        """
        if len(targets) == 0:
            return bool(self.get('enabled', 'true'))

        return self.name.lower() in targets

    def append(self, setting, custom_key=None):
        if not isinstance(setting, Setting):
            raise TypeError
        if custom_key is None:
            key = self.__prepare_key(setting.key)
        else:
            key = self.__prepare_key(custom_key)

        # Setting asserts key != "" for us
        self.contents[key] = setting

    def add_or_create_setting(self,
                              setting,
                              custom_key=None,
                              allow_appending=True):
        """
        Adds the value of the setting to an existing setting if there is
        already a setting  with the key. Otherwise creates a new setting.
        """
        if custom_key is None:
            key = setting.key
        else:
            key = custom_key

        if self.__contains__(key, ignore_defaults=True) and allow_appending:
            val = self[key]
            val.value = str(val._value) + '\n' + setting._value
            self.append(val, custom_key=key)
        else:
            self.append(setting, custom_key=key)

    @enforce_signature
    def __setitem__(self, key: str, value: (str, Setting)):
        """
        Creates a Setting object from the given value if needed and assigns the
        setting to the key:

        >>> section = Section('section_name')
        >>> section['key'] = 'value'
        >>> section['key'].value
        'value'

        :param key:   Argument whose value is to be set
        :param value: The value of the given key
        :return:      Returns nothing.
        """
        if isinstance(value, Setting):
            self.append(value, custom_key=key)
        else:  # It must be a string since signature is enforced
            self.append(Setting(key, value))

    def __iter__(self, ignore_defaults=False):
        joined = self.contents.copy()
        if self.defaults is not None and not ignore_defaults:
            # Since we only return the iterator of joined (which doesnt contain
            # values) it's ok to override values here
            joined.update(self.defaults.contents)

        return iter(joined)

    def __contains__(self, item, ignore_defaults=False):
        try:
            self.__getitem__(item, ignore_defaults)

            return True
        except IndexError:
            return False

    def __getitem__(self, item, ignore_defaults=False):
        key = self.__prepare_key(item)
        if key == '':
            raise IndexError('Empty keys are invalid.')

        res = copy.deepcopy(self.contents.get(key, None))
        if res is not None:
            if res.to_append and self.defaults and res.key in self.defaults:
                res.value = self.defaults[key]._value + ', ' + res._value
            res.to_append = False
            return res

        if self.defaults is None or ignore_defaults:
            raise IndexError('Required index is unavailable.')

        return self.defaults[key]

    def __str__(self):
        value_list = ', '.join(key + ' : ' + repr(str(self[key]))
                               for key in self.contents)
        return self.name + ' {' + value_list + '}'

    def get(self, key, default='', ignore_defaults=False):
        """
        Retrieves the item without raising an exception. If the item is not
        available an appropriate Setting will be generated from your provided
        default value.

        :param key:             The key of the setting to return.
        :param default:         The default value
        :param ignore_defaults: Whether or not to ignore the default section.
        :return:                The setting.
        """
        try:
            return self.__getitem__(key, ignore_defaults)
        except IndexError:
            return Setting(key, str(default))

    def copy(self):
        """
        :return: a deep copy of this object
        """
        result = copy.copy(self)
        result.contents = copy.deepcopy(self.contents)
        if self.defaults is not None:
            result.defaults = self.defaults.copy()

        return result

    def update(self, other_section, ignore_defaults=False):
        """
        Incorporates all keys and values from the other section into this one.
        Values from the other section override the ones from this one.

        Default values from the other section override the default values from
        this only.

        :param other_section:   Another Section
        :param ignore_defaults: If set to true, do not take default values from
                                other
        :return:                self
        """
        if not isinstance(other_section, Section):
            raise TypeError('other_section has to be a Section')

        self.contents.update(other_section.contents)

        if not ignore_defaults and other_section.defaults is not None:
            if self.defaults is None:
                self.defaults = other_section.defaults.copy()
            else:
                self.defaults.update(other_section.defaults)

        return self

    def update_setting(self,
                       key,
                       new_key=None,
                       new_value=None):
        """
        Updates a setting with new values.
        :param key:       The old key string.
        :param new_key:   The new key string.
        :param new_value: The new value for the setting
        """
        if new_key is not None:
            self.contents[key].key = new_key
            self.contents = update_ordered_dict_key(self.contents,
                                                    key,
                                                    new_key)
        if new_value is not None:
            if new_key is not None:
                self.contents[new_key].value = new_value
            else:
                self.contents[key].value = new_value

    def delete_setting(self, key):
        """
        Delete a setting
        :param key: The key of the setting to be deleted
        """
        del self.contents[key]

    def set_default_section(self, sections, section_name=None):
        """
        Find and set the defaults of a section from a dictionary of sections.
        The defaults are found on the basis of '.' in section names:

        >>> sections = {'all': Section('all')}
        >>> section = Section('all.python')
        >>> section.set_default_section(sections)
        >>> section.defaults.name
        'all'
        >>> section = Section('all.python.syntax')
        >>> section.set_default_section(sections)
        >>> section.defaults.name
        'all'

        This works case insensitive. The key of the sections dict
        is expected to be lowered though!

        >>> sections = {'c': Section('C'), 'cpp': Section('Cpp'),
        ...             'c.something': Section('C.something')}
        >>> section = Section('C.something')
        >>> section.set_default_section(sections)
        >>> section.defaults.name
        'C'
        >>> section = Section('C.SOMETHING.else')
        >>> section.set_default_section(sections)
        >>> section.defaults.name
        'C.something'
        >>> section = Section('Cpp.SOMETHING.else')
        >>> section.set_default_section(sections)
        >>> section.defaults.name
        'Cpp'

        :param sections:     A dictionary of sections.
        :param section_name: Optional section name argument to find the default
                             section for. If not given then use member section
                             name.
        """
        default_section = '.'.join(
            (section_name or self.name).split('.')[:-1]
        ).lower()

        if default_section:
            if default_section in sections:
                self.defaults = sections[default_section]
            else:
                self.set_default_section(sections, default_section)
        elif 'cli' in sections and self.name.lower() != 'cli':
            # CLI section is now default
            self.defaults = sections['cli']
